import numpy as np
import pandas as pd
import scipy.spatial as sp, scipy.cluster.hierarchy as hc
from sklearn.cluster import KMeans
from hdbscan import HDBSCAN

def calc_kernel_PCA(matrix, kernel = "precomputed") :
	from sklearn.decomposition import KernelPCA
	pca = KernelPCA(kernel = kernel)
	gram = 1 - (matrix / matrix.max())

	xy = pca.fit_transform(gram)

	return(xy)

def calculate_linkage(mtx, index_vals) :
	mtx_df = pd.DataFrame(mtx, index = index_vals, columns = index_vals)
	mtx_corr = mtx_df.corr()
	mtx_dism = 1 - mtx_corr
	linkage = hc.linkage(sp.distance.squareform(mtx_dism), method='average')

	return(linkage)

def cut_dist_tree(linkage, n_clusts) :
	hc.cut_tree(linkage, n_clusters = n_clust)

def kmeans(data, k = 30) :
	kmeans = KMeans(n_clusters = k, random_state=0).fit(data)

	return(kmeans.labels_)

def hdbscan(data, **kwargs) :
	clusterer = HDBSCAN(**kwargs)
	clusterer.fit(data)
	return(clusterer.labels_)

def tcrdiv(data, by_factor = None) :
	'''Calculate the TCRDiv as desribed by Dash et al.  
	Derived from TCRDist github: analyze_overlap_compute_simpsons.py'''
	div_dict = {}
	if by_factor :
		if by_factor not in data.obs.columns :
			raise ValueError("{by_fac} not in .obs".format(by_fac = by_factor))
		unique_factors = set(data.obs[by_factor])

		for fac in unique_factors :
			data_sub = data[data.obs[by_factor] == fac]
			div = _calc_tcrdiv(data_sub.X)
			div_dict[fac] = div

	else :
		div = _calc_tcrdiv(data.X)
		div_dict["all"] = div

	return(div_dict)

# Modified such that it dynamically calculates the standard deviationof the input matrix
def _calc_tcrdiv(matrix) :
	overlap_sum = 0
	total_sum = 0
	
	sd = np.std(matrix)
	for i in range(matrix.shape[0]) :
		for j in range(i + 1, matrix.shape[0]) :
			dist = matrix[i,j]
			if dist == 0 :
				overlap_sum += 1
			else :
				overlap_sum += np.exp(-1 * (dist/sd)**2)
			total_sum += 1

	p0 = overlap_sum / total_sum
	diversity = 1 / p0
	return(diversity.item())




'''Compute distances of all V genes and store for reference'''
import json
import distance

def compute_all_v_region_distances(organism):
    with open("db/{organism}_cdrs_for_10X.json".format(organism = organism), "r") as f:
        cdr_dict = json.load(f)

    # Get all unique V gene names
    keys = [list(cdr_dict[key].keys()) for key in cdr_dict.keys()]
    assert keys[0] == keys[1] == keys[2]
    keys = keys[0]

    # Collect all of the information for each V gene
    v_gene_dict = {}
    for v_gene in keys :
        # Get all of the cdrs for each V gene
        cdrs = [cdr_dict[cdr][v_gene] for cdr in cdr_dict.keys()]
        
        v_gene_dict[v_gene] = cdrs
    v_comp_dict = {}
    # Get distance of possible V gene matches
    for chain in ["A", "B"] :
        # Get V genes from a single chain
        seqs = {key : v_gene_dict[key] for key in v_gene_dict.keys() if chain in key}
        for v_1 in seqs.keys() :
            for v_2 in seqs.keys() :
                aligned = [distance.align_cdrs(seq1, seq2) for seq1, seq2 in zip(seqs[v_1], seqs[v_2])]
                dists = [distance.blosum_sequence_distance(seq1, seq2, gap_penalty = 4) for seq1, seq2 in aligned]
                total_dist = sum(dists)
                v_comp_dict["_".join([v_1, v_2])] = total_dist
    with open("db/{organism}_v_gene_dists.json".format(organism = organism), "w") as f :
            json.dump(v_comp_dict, f, indent = 4)
            f.close()

if __name__ == "__main__" :
    compute_all_v_region_distances("human")
    compute_all_v_region_distances("mouse")


import seaborn as sns
import pandas as pd
from . import useful
from .tcrdist import stats as stat
import matplotlib.pyplot as plt
import numpy as np	
import logomaker
from collections import Counter

def plot_distance_matrix(
	data,
	row_colors : str = None,
	**kwargs) :
	
	if row_colors :
		if row_colors not in  data.obs.columns :
			raise ValueError("{row_col} not in .obs".format(row_col = row_colors))
		# Get the categories
		if not pd.api.types.is_categorical(data.obs[row_colors]) :
			data.obs[row_colors] = data.obs[row_colors].astype("category")
		categories = data.obs[row_colors].cat.categories

		if not "{row_col}_colors".format(row_col = row_colors) in data.uns :

			palette = useful.godsnot_102[:len(categories)]
			data.uns["{row_col}_colors".format(row_col = row_colors)] = palette

		col_dict = dict(zip(categories, data.uns["{row_col}_colors".format(row_col = row_colors)]))

		row_colors = [col_dict[gene] for gene in data.obs[row_colors]]

	if not "linkage" in data.uns :
		linkage = stat.calculate_linkage(data.X, data.obs.index.values)
		data.uns["linkage"] = linkage
	else :
		linkage = data.uns["linkage"]
	plot_df = pd.DataFrame(data.X, index = data.obs.index.values)
	plot = sns.clustermap(plot_df, row_linkage=linkage, col_linkage=linkage, row_colors = row_colors, **kwargs)
	return(plot)


def cluster_logo_plot(adata, obs_col, obs_val, lengths = "all") :
	length_args = ["all", "dominant"]
	if lengths not in length_args :
		raise ValueError("length argument must be one of %s" % length_args)

	# Lets take an example cluster
	logo_clust = adata.obs[adata.obs[obs_col] == obs_val][["TRA_cdr3", "TRB_cdr3", "TRA_cdr3_length", "TRB_cdr3_length"]]

	# Figure out the dominant lengths of the clusters
	if lengths == "dominant" :
		num_alpha_lengths = 1
		num_beta_lengths = 1
	else :
		num_beta_lengths = len(set(logo_clust["TRB_cdr3_length"]))
		num_alpha_lengths = len(set(logo_clust["TRA_cdr3_length"]))

	figRows = max([num_beta_lengths, num_alpha_lengths])

	# NEED TO FIGURE OUT HOW TO MAKE A GOOD FIGSIZE CALCULATION
	fig, ax = plt.subplots(nrows = figRows, ncols = 2, figsize = (10 * figRows, 3 * figRows))

	for num, seqtype in enumerate(["TRA", "TRB"]) :
		seq_df = logo_clust[["{seqtype}_cdr3".format(seqtype = seqtype), "{seqtype}_cdr3_length".format(seqtype = seqtype)]]

		if lengths == "dominant" :
			chain_lengths = [seq_df["{seqtype}_cdr3_length".format(seqtype = seqtype)].value_counts().idxmax()]
		else :
			chain_lengths = sorted(set(seq_df["{seqtype}_cdr3_length".format(seqtype = seqtype)]))
		for row, seqlen in enumerate(chain_lengths) :
			# Get the seqs
			logo_seqs = seq_df[seq_df["{seqtype}_cdr3_length".format(seqtype = seqtype)] == seqlen]["{seqtype}_cdr3".format(seqtype = seqtype)]
			# Concatenate and determine all used AA in seqs
			unique_AA = set("".join(seq for seq in logo_seqs))
			# Probability matrix
			prob_df = pd.DataFrame(index = range(seqlen), columns = unique_AA)

			for indx in range(len(logo_seqs[0])) :
				# Get the letter at the position for each seq
				AAs = [seq[indx] for seq in logo_seqs]

				# Calculate probabilities
				prob_dict = dict(Counter(AAs))
				for key, val in prob_dict.items() :
					prob = val / len(AAs)
					prob_df.loc[indx, key] = prob
			prob_df = prob_df.fillna(0)
			prob_df.sum(axis = 1)

			if figRows == 1 :
				logomaker.Logo(prob_df, ax = ax[num],
								width=.8,
								vpad=.05,
								fade_probabilities=True,
								stack_order='small_on_top',
								color_scheme='dodgerblue',
								font_name='Rosewood Std'
							   )
				ax[num].set_title("Number of seqs: {seqlen}".format(seqlen = len(logo_seqs)), {"fontsize" : 10, "fontweight" : "bold"})

				# Add additional title
				# Get the center of the plot
				center = seqlen / 1.75
				height = 1.1 + (figRows / 15)
				ax[num].text(center, height, "{seqtype} CDR3".format(seqtype = seqtype),{"fontsize" : 15, "fontweight" : "bold"} ,
								  horizontalalignment = "right")
				continue
			else :
				logomaker.Logo(prob_df, ax = ax[row, num],
								width=.8,
								vpad=.05,
								fade_probabilities=True,
								stack_order='small_on_top',
								color_scheme='dodgerblue',
								font_name='Rosewood Std'
							   )
				ax[row, num].set_title("Number of seqs: {seqlen}".format(seqlen = len(logo_seqs)), {"fontsize" : 10, "fontweight" : "bold"})
			# If the first of either alpha or beta, add additional title
				if row == 0 :
					# Get the center of the plot
					center = (seqlen + .75) / 2
					height = 1 + (figRows / 15)
					ax[row, num].text(center, height, "{seqtype} CDR3".format(seqtype = seqtype),{"fontsize" : 15, "fontweight" : "bold"} ,
									  horizontalalignment = "right")
	fig.tight_layout()

	# Determine which chain has more seq lengths
	if num_beta_lengths > num_alpha_lengths :
		# get rid of excess alpha plots
		for noLogo in range(num_alpha_lengths, num_beta_lengths) :
			ax[noLogo, 0].axis("off")
	elif num_beta_lengths < num_alpha_lengths :
		# Get rid of excess beta plots
		for noLogo in range(num_beta_lengths, num_alpha_lengths) :
			ax[noLogo, 1].axis("off")

	return(fig)


import pandas as pd
import json
import os
from .. import plot
from .cython import seq_dist
from . import stats
from anndata import AnnData
import numpy as np
import time

class tcrdist_dataset(AnnData) :
    def __init__(self, seq_data, organism, only_betas = False) :
        super(tcrdist_dataset, self).__init__(shape = (seq_data.shape[0], seq_data.shape[0]))
        self.obs = seq_data
        self.obs.columns = self.obs.columns.str.replace(" ", "_")
        self.obs.index = self.obs.index.map(str)
        self.uns["species"] = organism
        self.uns["only_betas"] = only_betas

        if organism not in ["human", "mouse"] :
            raise(ValueError("organism must be either human or mouse"))
        with open(os.path.join(os.path.dirname(__file__), "db", "{organism}_v_gene_dists.json".format(organism = organism)), "r") as f:
            self.uns["v_gene_dists"] = json.load(f)
        for chain in ["TRA_cdr3", "TRB_cdr3"] :
            try :
                self.obs["_".join([chain, "trimmed"])] = self.trim_cdr3s(seq_data[chain])
            except :
                continue

        # Get information about CDR3 length
        for chain in ["TRA_cdr3", "TRB_cdr3"] :
            try :
                self.obs["_".join([chain, "length"])] = self.obs.apply(lambda df : len(df[chain]), axis = 1)
            except :
                continue

    def trim_cdr3s(self, seqs, front_trim = 3, back_trim = 2) :
        '''tcrdist calculation used a trimmed CDR3:
         "trimmed CDR3’ defined as starting with the 3rd position after the C104 
         and terminating with the 2nd position before F118 (5 fewer residues than the full CDR3"
        Dash et al. 2017.
         '''
        trimmed_seqs = [seq[front_trim:-back_trim] for seq in seqs]
        return(trimmed_seqs)

    def calc_distance_legacy(self, seq1_info, seq2_info) :
        
        # Calculate the distance of the CDR3s
        aligned_A = seq_dist.align_cdrs(seq1_info["TRA_cdr3_trimmed"], seq2_info["TRA_cdr3_trimmed"])
        cdr3_dist_A = seq_dist.blosum_sequence_distance(aligned_A[0], aligned_A[1], gap_penalty = 12, cdr3 = True)

        aligned_B = seq_dist.align_cdrs(seq1_info["TRB_cdr3_trimmed"], seq2_info["TRB_cdr3_trimmed"])
        cdr3_dist_B = seq_dist.blosum_sequence_distance(aligned_B[0], aligned_B[1], gap_penalty = 12, cdr3 = True)

        # Get the distances of the V genes
        v_dist_A = self.uns["v_gene_dists"]["_".join([seq1_info["TRA_v_gene"], seq2_info["TRA_v_gene"]])]
        v_dist_B = self.uns["v_gene_dists"]["_".join([seq1_info["TRB_v_gene"], seq2_info["TRB_v_gene"]])]
        tot_dist = sum([cdr3_dist_A, cdr3_dist_B, v_dist_A, v_dist_B])

        return(tot_dist)

    def create_tcrdist_matrix_legacy(self, barcodes) :
        import time
        start = time.time()
        mtx = np.zeros((len(barcodes), len(barcodes)))

        for code1 in range(len(barcodes)) :
            for code2 in range(code1 + 1, len(barcodes)) :
                data1 = self.collect_info(barcodes[code1])
                data2 = self.collect_info(barcodes[code2])

                dist = self.calc_distance_legacy(data1, data2)
                mtx[code1, code2] = dist
                mtx[code2, code1] = dist
        end = time.time()
        print("time for mtx: %s seconds"%(end - start))
        self.X = mtx
        return(mtx)
    
    def create_tcrdist_matrix_nw(self) :
            start = time.time()
            mtx = seq_dist.create_tcrdist_matrix(self.obs, v_scores = self.uns["v_gene_dists"], only_betas = self.uns["only_betas"])
            end = time.time()
            print("time for mtx: %s seconds"%(end - start))
            print(mtx)
            self.X = mtx

    def collect_info(self, barcode) :
        data = self.obs.loc[barcode]
        info = {
        "TRA_cdr3_trimmed" : data["TRA_cdr3_trimmed"],
        "TRB_cdr3_trimmed" : data["TRB_cdr3_trimmed"],
        "TRA_v_gene" : data["TRA_v_gene"],
        "TRB_v_gene" : data["TRB_v_gene"]
        }
        return(info)

    def kernel_pca(self, kernel = "precomputed") :
        pca = stats.calc_kernel_PCA(self.X, kernel = kernel)
        self.obsm["k_pca"] = pca

    def calc_umap(self, return_coords = False) :
        import umap
        # Make sure kernel-PCA has been calculated
        # if not "k_pca" in self.obsm :
        #     raise ValueError("no PCA in object.  Please run kernel_pca first!")
        
        reducer = umap.UMAP()
        embedding = reducer.fit_transform(self.X)

        self.obsm["X_umap"] = embedding

        if return_coords :
            return(embedding)

    def hdbscan_cluster(self, return_results = False, coord = "umap", n_coord = 2, **kwargs) :
        embedding = "_".join(["X", coord])
        if not embedding in self.obsm :
            raise ValueError("No {coord} in object.  Please run {coord} first!".format(coord = coord))
            
        clust = stats.hdbscan(self.obsm[embedding][:,0:n_coord], **kwargs)
        # Need to add one because pandas gets confused with categorical data < 1, make -1 "unassigned"
        clust = [str(num + 1) if num != -1 else "unassigned" for num in clust]
        self.obs["hdbscan_clust"] = clust

        if return_results :
            return(clust)

def tcrdist_dataset_from_anndata(adata, organism, obs = []) :
    '''Function to initialize an object of 'tcrdist_dataset' from an anndata that has been processed via pegasus'''

    adata = adata[(adata.obs["TRB_cdr3"] != "") & (adata.obs["TRA_cdr3"] != "") & (adata.obs["TRA_v_gene"] != "") & (adata.obs["TRB_v_gene"] != "")]
    
    # Get the needed columns
    need_cols = ["TRA_cdr3", "TRB_cdr3", "TRA_v_gene", "TRB_v_gene"] + obs

    needed_data = adata.obs[need_cols]
    # Removes categorical datatype if it exists
    needed_data = pd.DataFrame(needed_data, dtype = "object")
    dset = tcrdist_dataset(seq_data = needed_data, organism = organism)

    return(dset)

def tcrdist_dataset_from_vdjdb(tsv_path, organism, only_betas = False, cell_id = "complex.id", chain_col = "Gene", cdr3_col = "CDR3", v_gene_col = "V",
    other_cols = []) :
    '''Function to initialize a tcrdist object from data downloaded from vdjdb (vdjdb.cdr3.net)'''
    data = pd.read_csv(tsv_path, sep = "\t")

    # Get just the data we need
    data = data[[cell_id, chain_col, cdr3_col, v_gene_col] + other_cols]
    # Adjust the formatting of the V genes
    data["V"] = [str(gene).replace("*", "-") for gene in data["V"]]
    # remove additional allele information(not needed)
    data["V"] = [gene[:-3] for gene in data["V"]]
    # Get rid of the slashes  in some of the names
    data["V"] = [gene.replace("/", "") for gene in data["V"]]

    if only_betas :
        needed_data = data[data["Gene"] == "TRB"]
        needed_data = needed_data.rename(columns = {"CDR3" : "TRB_cdr3", "V" : "TRB_v_gene"})
        needed_data = needed_data.drop(["Gene", "complex.id"], axis = 1)

    else :
    # Seperate the alpha and beta data
        df_dict = {}
        for chain in ["TRA", "TRB"] :
            df = data[data["Gene"] == chain]
            if chain == "TRA" :
                df = df.drop(other_cols, axis = 1)
            df = df.rename(columns = {"CDR3" : "{chain}_cdr3".format(chain = chain), "V" : "{chain}_v_gene".format(chain = chain)})
            df = df.drop(["Gene"], axis = 1)
            df = df.set_index("complex.id")

            df_dict[chain] = df
        needed_data = df_dict["TRA"].join(df_dict["TRB"], on = "complex.id")

        # Remove the TCRs with missing information
        needed_data = needed_data[(needed_data["TRB_cdr3"] != "") &
        (needed_data["TRA_cdr3"] != "") & 
        (needed_data["TRA_v_gene"] != "") & 
        (needed_data["TRB_v_gene"] != "")]
        needed_data = needed_data.dropna()

    dset = tcrdist_dataset(seq_data = needed_data, organism = organism)
    return(dset)

def tcrdist_dataset_from_csv(csv_path, organism) :
    # needs to be done
    return(0)

def tcrdist_dataset_from_df(df, organism, only_betas = False, from_adaptive = False, cdr3a_col = "TRA_crd3", cdr3b_col = "TRB_cdr3", Va_col = "TRA_v_gene", Vb_col = "TRB_v_gene") :
    if only_betas :
        needed_data = df
        needed_data = needed_data.rename(columns = {cdr3b_col : "TRB_cdr3", Vb_col : "TRB_v_gene"})
    else :
        needed_data = df
        needed_data = needed_data.rename(columns = {cdr3a_col : "TRA_cdr3", cdr3b_col : "TRB_cdr3", Va_col: "TRA_v_gene", Vb_col : "TRB_v_gene"})

    if from_adaptive :
        needed_data["TRB_v_gene"] = [gene if gene != "TRBV9-1" else "TRBV9" for gene in needed_data["TRB_v_gene"]]
        needed_data["TRB_v_gene"] = [gene if gene != "TRBV26-1" else "TRBV26" for gene in needed_data["TRB_v_gene"]]
        needed_data["TRB_v_gene"] = [gene if gene != "TRBV1-1" else "TRBV1" for gene in needed_data["TRB_v_gene"]]

    dset = tcrdist_dataset(seq_data = needed_data, organism = organism, only_betas = only_betas)
    return(dset)


from setuptools import setup, find_packages
from codecs import open
from os import path
from Cython.Build import cythonize
from distutils.core import setup, Extension
import numpy

here = path.abspath(path.dirname(__file__))

np_include = numpy.get_include()
include_dirs = [np_include, "fast_tcrdist/tcrdist/cython"]

extensions = [Extension('fast_tcrdist.tcrdist.cython.seq_dist', ['fast_tcrdist/tcrdist/cython/seq_dist.pyx']),
Extension("fast_tcrdist.tcrdist.cython.cnwalign", ["fast_tcrdist/tcrdist/cython/cnwalign.pyx"], include_dirs = include_dirs)]

setup(
    name='fast_tcrdist',

    version='0.0.1',

    description='Optimized TCRDist calculation for TCR repertoire data analysis',
    long_description='',

    url='',

    author='Neal Smith',
    author_email='nsmith19@mgh.harvard.edu',

    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Natural Language :: English'
    ],

    install_requires=["pandas", "numpy", "distance", "cython", "hdbscan", "logomaker", "anndata", "scanpy"],

    packages=find_packages(),

    ext_modules = cythonize(extensions), include_dirs = include_dirs
)



import pandas as pd 
import numpy as np
import h5py
import time
import anndata
import scanpy as sc
def aggregate_cellranger_tcr(
	sample_sheet,
	save_location,
	chains = ["TRA", "TRB"],
	chain_vars = ["cdr3", "v_gene"]) :
	
	# Read in the sample info
	samp_info = pd.read_csv(sample_sheet)

	# Make sure columns are correct
	if not all([col in samp_info.columns for col in ["h5_location", "tcr_location"]]) :
		raise(ValueError( "sample sheet needs 2 columns : h5_location and tcr_location"))

	# Make a dict of chain/dtype combos
	# chain_dict = {key: chain_vars for key in chains}

	samp_dict = dict(zip(samp_info["h5_location"], samp_info["tcr_location"]))

	anndatas = []
	for gex, tcr in samp_dict.items() :
		gex_data = sc.read_10x_h5(gex)
		gex_data.var_names_make_unique()
		tcr_data = pd.read_csv(tcr)

		tcr_data = tcr_data[(tcr_data["productive"] == "True") & (tcr_data["high_confidence"] == True)]

		for dtype in chain_vars :
			for chain in chains :
				chain_data = tcr_data[tcr_data["chain"] == chain]
				info_dict = dict(zip(chain_data["barcode"], chain_data[dtype]))

				gex_data.obs["_".join([chain, dtype])] = [info_dict[code] if code in info_dict.keys() else "" for code in gex_data.obs.index]
		anndatas.append(gex_data)
	if len(anndatas) > 1 :
		all_data = anndatas[0].concatenate(anndatas[1:])
	else :
		all_data = anndatas[0]

	# Save the data
	all_data.write_h5ad(save_location)